/*

xspectrum.c

nice little program for visalizing spectrum of an audio signal under X11.

(C) Philippe Strauss <philippe.strauss@urbanet.ch>, 1995, 1996.

this program is licensed under the term of the GNU General Public License 2,
which is included in the file GPL2.

*/

/*
 * This is jxspectrum, a modified version of xspectrum.
 */

/* Changes by JF:
 * 970918: Removed all display code and EZ specific stuff
 * 970819: Removed unneccessery subroutines 
 *         Many changes to DoFFT() and main()
 * 970920: Removed some static identifiers 
 *         Changed SetSF() to take an argument
 * 970921: Added functions SetWinKaiser() ,SetWinBlackman() and SetOpmode()
 *         Changed Quit()
 */

#include "gui.h"
#include "xspectrum.h"
#include "mydsp.h"

#ifdef DEBUG
#undef DEBUG
#endif

/*****************************************************************************/
/* functions code */

//New functions by JF

void SetWinKaiser()
{
    MyInitWinKaiser (sPrivate.pfWindow, sPrivate.iBeta, sPrivate.iN);
}

void SetWinBlackman()
{
    MyInitWinBlackman (sPrivate.pfWindow,sPrivate.iN);
}

void SetWinRectangle()
{
    MyInitWinRectangle (sPrivate.pfWindow,sPrivate.iN);
}

void SetOpmode (int mode)
{
    sPrivate.Opmode=mode;
}

/*
QuitOnError : Do a simple exit on error, print the errno value.
*/
static void QuitOnError (const char * pchrErrorMessage)
{
    int iZero = 0;

    perror (pchrErrorMessage);

    /* toggle enable bit to zero */

    if (ioctl (sPrivate.iFdDSP, SOUND_PCM_SETTRIGGER, &iZero))
        perror("Error on ioctl SNDCTL_DSP_SETTRIGGER");

    exit (1);
}

/*
AllocCmplxRect : Allocate memory for the arrays in a CmplxRect structure
*/
static void AllocCmplxRect (tsCmplxRect *psCmplxRect, tInt iN)
{
    psCmplxRect->pfReal = (ptFloat) calloc (iN, (size_t) sizeof (tFloat));
    psCmplxRect->pfImag = (ptFloat) calloc (iN, (size_t) sizeof (tFloat));

    if ((psCmplxRect->pfReal == NULL) || (psCmplxRect->pfImag == NULL))
        QuitOnError ("Error: Unable to allocate memory for CmplxRect");
}

static void FreeCmplxRect (tsCmplxRect *psCmplxRect)
{
    free (psCmplxRect->pfReal);
    free (psCmplxRect->pfImag);
}

/*
SoundRead_S16_LE : Read from soundcard, pBuffer has been mapped as the DMA
                   buffer. Data readed from the soundcard must be in
                   signed 16 bits Little Endian format.
*/
static void SoundRead_S16_LE ()
{
    static int iBuffSize, iBeginDMA, iCurrentDMA, iIdxDest, iIdxDMA;
    static count_info sCountInfo;
    static volatile int16 *pLocalPtr;
    static int iPrevCntPtr = 0;
#   ifdef DEBUG
    static int iDenom;
    static float fFps;
#   endif

    iBuffSize = sPrivate.iBuffLen / sizeof (int16);
    /* convenient way to convert data format. pBuffer is of type caddr_t */
    /* which is a typedef of char *, data returned by most soundcards is in */
    /* 16 bits little endian, which hoppefully is the short on i386 linux */
    pLocalPtr = (volatile int16 *) sPrivate.pBuffer;
    
    iPrevCntPtr = sCountInfo.ptr;

    if (ioctl (sPrivate.iFdDSP, SNDCTL_DSP_GETIPTR, &sCountInfo) != 0)
        QuitOnError ("Error on ioctl SNDCTL_DSP_GETIPTR");

#   ifdef DEBUG
    /* calculate the number of frame per second */
    iDenom = (sCountInfo.ptr - iPrevCntPtr + sPrivate.iBuffLen)
             % sPrivate.iBuffLen;
    if (iDenom != 0)
        fFps = (sPrivate.iSpeed * sizeof (int16)) / (float) iDenom;
    else
       fFps = 0.0;
 
    printf ("\rDebug:\tDMA pointer %05d, Refresh rate %3.1f fps",
            sCountInfo.ptr, fFps);
    fflush (stdout);
#   endif /* DEBUG */

/*  iCurrentDMA = (sCountInfo.ptr - 1) / sizeof (int16);*/
    iCurrentDMA = (sCountInfo.ptr) / sizeof (int16);

    /* circular adressing of DMA buffer */
    iBeginDMA   = (iCurrentDMA + iBuffSize - sPrivate.iN) % iBuffSize;

    for (iIdxDest = 0, iIdxDMA = iBeginDMA;
         iIdxDest <= sPrivate.iN;
         ++iIdxDest, iIdxDMA = (iIdxDMA + 1) % iBuffSize)
    {
        sPrivate.sFFT.pfReal [iIdxDest] = pLocalPtr [iIdxDMA] / 32768.0;
        sPrivate.sFFT.pfImag [iIdxDest] = 0.0;
    }
}

/*
Initialize : Top-level initialization function.
*/
static void Initialize ()
{
    sPrivate.iSpeed      = 40000;
    sPrivate.iN          = NPOINTS;
    sPrivate.iOrder      = ORDER;
    sPrivate.iNChannels  = 1;
    sPrivate.iFormat     = AFMT_S16_LE;
    sPrivate.iBeta       = 5;
    sPrivate.uiWidth     = sPrivate.iN / 2; /* width of the chart window */
    sPrivate.uiHeight    = CHART_H;  /* height of the chart window */
    sPrivate.fDB2Pixel   = DB2PIXEL;
    sPrivate.fRefLevel   = REF_LEVEL;
    sPrivate.iBuffLen    = sizeof (int16) * sPrivate.iN;
    sPrivate.iRun        = 1;
    sPrivate.Opmode      = 0;
    
#   ifdef DEBUG
    printf ("Debug:\tdB2Pixel = %f\n", sPrivate.fDB2Pixel);
    printf ("Debug:\tRefLevel = %f\n", sPrivate.fRefLevel);
#   endif

    InitMemAlloc    ();
    InitSoundDevice ();
    InitFFT         ();
}

/*
InitMemAlloc : do all  memory allocation needed.
*/
static void InitMemAlloc ()
{
    AllocCmplxRect (&(sPrivate.sFFT), sPrivate.iN);
    AllocCmplxRect (&(sPrivate.sTwidTable), sPrivate.iN);

    sPrivate.piSwapTable = (ptInt) calloc (sPrivate.iN,
                                           (size_t) sizeof (tInt));
    if (sPrivate.piSwapTable == NULL)
        QuitOnError ("Cannot allocate memory for IntArray");

    sPrivate.pfWindow = (ptFloat) calloc (sPrivate.iN,
                                          (size_t) sizeof (tFloat));
    if (sPrivate.pfWindow == NULL)
        QuitOnError ("Cannot allocate memory for RealArray");

    sPrivate.pfMag = (ptFloat) calloc (sPrivate.iN, (size_t) sizeof (tFloat));
    if (sPrivate.pfMag == NULL)
        QuitOnError ("Cannot allocate memory for RealArray");

    /*sPrivate.pPoints =
        (XPoint *) calloc (sPrivate.iN / 2, (size_t) sizeof (XPoint));
    if (sPrivate.pPoints == NULL)
        QuitOnError ("Cannot allocate memory for Points");
 
    sPrivate.pGrid =
        (XSegment *) calloc ((GRID_H + GRID_V + 2),
                             (size_t) sizeof (XSegment));
    if (sPrivate.pGrid == NULL)
        QuitOnError ("Cannot allocate memory for Grid");*/

} /* InitMemAlloc */

/*
InitSoundDevice : deal with the kernel for opening /dev/dsp
*/
static void InitSoundDevice ()
{
    audio_buf_info sBuffInfo;

    if ((sPrivate.iFdDSP = open ("/dev/dsp", O_RDONLY)) == -1)
        QuitOnError ("Cannot open /dev/dsp");

    /* set number of channels */
    if (ioctl (sPrivate.iFdDSP, SNDCTL_DSP_CHANNELS,
               &(sPrivate.iNChannels)))
        QuitOnError ("Error on ioctl SNDCTL_DSP_CHANNELS");

    /* set the format */
    if (ioctl (sPrivate.iFdDSP, SNDCTL_DSP_SETFMT, &(sPrivate.iFormat)))
        QuitOnError ("Error on ioctl SNDCTL_DSP_SETFMT");
    if (sPrivate.iFormat != AFMT_S16_LE)
        QuitOnError ("What shitty hardware are you using :)\nI want 16 bits samples for a good S/N ratio, sorry");

    /* set the sampling rate */
    if (ioctl (sPrivate.iFdDSP, SNDCTL_DSP_SPEED, &(sPrivate.iSpeed)))
        QuitOnError ("Error on ioctl SNDCTL_DSP_SPEED");

#   ifdef DEBUG
    /* write on stdout how the sound driver has done things for us */
    printf ("Debug:\tNumber of channels is %d\n", sPrivate.iNChannels);
    printf ("Debug:\tSample rate is %d Hz\n", sPrivate.iSpeed);
    printf ("Debug:\tFormat is AFMT_S16_LE\n");
#   endif

    /* query capabilites */
    if (ioctl (sPrivate.iFdDSP, SNDCTL_DSP_GETCAPS,
               &(sPrivate.iCaps)) != 0)
        QuitOnError ("Error on ioctl SNDCTL_DSP_GETCAPS");

#   ifdef DEBUG
    if (sPrivate.iCaps & DSP_CAP_DUPLEX)
        printf ("Debug:\tSound device support full duplex mode\n");

    if (sPrivate.iCaps & DSP_CAP_REALTIME)
        printf ("Debug:\tSound device has realtime capabilities\n");

    if (sPrivate.iCaps & DSP_CAP_MMAP)
        printf ("Debug:\tSound device support memory mapping I/O\n");

    if (sPrivate.iCaps & DSP_CAP_TRIGGER)
        printf ("Debug:\tSound device support triggering\n");
#   endif /* DEBUG */

    /* we must obtain mmaping */
    if ((sPrivate.iCaps & DSP_CAP_MMAP) &&
        (sPrivate.iCaps & DSP_CAP_TRIGGER))
    {
        if (ioctl (sPrivate.iFdDSP, SNDCTL_DSP_GETISPACE,
                   &sBuffInfo) != 0)
            QuitOnError ("Error on ioctl SNDCTL_DSP_GETISPACE");

        sPrivate.iBuffLen = sBuffInfo.fragsize * sBuffInfo.fragstotal; 

#       ifdef DEBUG
        printf ("Debug:\tDMA Buffer size is %d bytes\n", sBuffInfo.fragsize
                                                       * sBuffInfo.fragstotal);
        printf ("Debug:\tNumber of available fragments is %d\n",
                sBuffInfo.fragments);
        printf ("Debug:\tNumber of fragments allocated is %d\n",
                sBuffInfo.fragstotal);
        printf ("Debug:\tSize of a fragment is %d bytes\n",
                sBuffInfo.fragsize);
        printf ("Debug:\tAvailable size is %d bytes\n", sBuffInfo.bytes);
        printf ("Debug:\tWe use %d bytes\n", sPrivate.iBuffLen);
#       endif /* DEBUG */

        /* map the DMA buffer in the proc vm */
        if ((sPrivate.pBuffer
            = (caddr_t) mmap (
                     NULL, /* hint adress for kernel, free if NULL */
                     sPrivate.iBuffLen,   /* length of buffer */
                     PROT_READ,           /* reading access only */
                     MAP_FILE|MAP_SHARED, /* map to a file or a char device */
                     sPrivate.iFdDSP,     /* file descriptor */
                     0                    /* offset inside of file */)
            ) == (void *) -1)
            QuitOnError ("Cannot mmap, Arrgh");

#       ifdef DEBUG
        printf ("Debug:\tDMA buffer adress is %x\n",
                (unsigned int) sPrivate.pBuffer);
#       endif /* DEBUG */

    }
    else
        QuitOnError ("Sound device does not support memory mapping I/O");
} /* InitSoundDevice */

static void CloseSoundDevice ()
{
    StopRecording ();
    munmap        (sPrivate.pBuffer, sPrivate.iBuffLen);
    close         (sPrivate.iFdDSP);
}

void SetSF (int speed)
{
    StopRecording ();
 
    sPrivate.iSpeed=speed;
    
    /* set the sampling rate */
    if (ioctl (sPrivate.iFdDSP, SNDCTL_DSP_SPEED, &(sPrivate.iSpeed)))
        QuitOnError ("Error on ioctl SNDCTL_DSP_SPEED");

    StartRecording ();
}

/*
InitFFT : initalize window, twiddle table and swap table.
*/
static void InitFFT ()
{
    /* initialize pfWindow with a Kaiser window in it */
    //MyInitWinKaiser (sPrivate.pfWindow, sPrivate.iBeta, sPrivate.iN);
    MyInitWinBlackman (sPrivate.pfWindow, sPrivate.iN);
    
    /* initialize sTwidTable with the twiddle factors */
    InitTwidTable (sPrivate.sTwidTable, DIRECT, sPrivate.iN);

    /* initialize piSwapTable with reversed digit for reordering fft coeffs */
#   ifdef RADIX4
    InitR4SwapTable (sPrivate.piSwapTable, sPrivate.iOrder, sPrivate.iN);
#   endif
#   ifdef RADIX2
    InitR2SwapTable (sPrivate.piSwapTable, sPrivate.iN);
#   endif
}

void StartRecording ()
{
    int iEnableInput = PCM_ENABLE_INPUT;

    /* toggle bit to zero */
    StopRecording ();

    /* toggle enable bit to one, start recording */
    if (ioctl (sPrivate.iFdDSP, SOUND_PCM_SETTRIGGER, &iEnableInput))
        QuitOnError ("Error on ioctl SNDCTL_DSP_SETTRIGGER");
}

void StopRecording ()
{
    int iZero = 0;

    /* toggle enable bit to zero */
    if (ioctl (sPrivate.iFdDSP, SOUND_PCM_SETTRIGGER, &iZero))
        QuitOnError ("Error on ioctl SNDCTL_DSP_SETTRIGGER");
}

/* main processing task as a work procedure */
void DoFFT (PlotPoint *points)
{
    static int iIdx;
    float tmpy;
    float tmpx= ((float) sPrivate.iSpeed)/((float) sPrivate.iN);
    
    SoundRead_S16_LE ();
    WindowingCR (sPrivate.sFFT, sPrivate.pfWindow, sPrivate.iN);
#   ifdef RADIX4
    R4FFTdif    (sPrivate.sFFT, sPrivate.sTwidTable, sPrivate.iOrder,
                 sPrivate.iN);
#   endif
#   ifdef RADIX2
    R2FFTdif    (sPrivate.sFFT, sPrivate.sTwidTable, sPrivate.iOrder,
                 sPrivate.iN);
#   endif
    SwapFFTC    (sPrivate.sFFT, sPrivate.piSwapTable, sPrivate.iN);

    Magnitude2  (sPrivate.sFFT, sPrivate.pfMag, sPrivate.iN);

    for (iIdx = 1; iIdx < (int) sPrivate.uiWidth; ++ iIdx)
    {
        /* must be non-zero otherwise log10 go -infinity ! */
        if (sPrivate.pfMag [iIdx] > 0.0)
        {
	    tmpy = 10*(log10(sPrivate.pfMag [iIdx]) - sPrivate.fRefLevel)
	      + sPrivate.fRefLevel;
	}
        else
	  tmpy = -200;
	
	points[iIdx-1].x=iIdx*tmpx;
	
	if(sPrivate.Opmode)
	{
	    if(tmpy>points[iIdx-1].y)
	      points[iIdx-1].y=tmpy;
	}
	else
	points[iIdx-1].y=tmpy;
    }
} /* proc */

void Quit ()
{    
    CloseSoundDevice ();
    FreeCmplxRect    (&(sPrivate.sFFT));
    FreeCmplxRect    (&(sPrivate.sTwidTable));
    free             (sPrivate.piSwapTable);
    free             (sPrivate.pfWindow);
    free             (sPrivate.pfMag);
#   ifdef DEBUG
    /* if we outputed some debug info in SoundRead_S16_LE, we need this */
    fflush           (stdout);
    printf           ("\n");
#   endif
}

/*****************************************************************************/

int main (int argc, char **argv)
{
    Initialize ();

    StartRecording ();

    QApplication app( argc, argv );
    Gui gui;
    
    app.setMainWidget ( &gui);
    gui.show();
        
    return app.exec();
}
